import argparse
import logging

from colorama import Fore

from .db import get_default_db, save_db


class Client(object):
    global_arguments = {('app', 'application'), ('pg', 'DB instance')}

    def __init__(self):
        self.modules = []
        self.parser = argparse.ArgumentParser()
        self.parser.add_argument(
            f'--log',
            help='Log level. Default: WARNING',
            choices=('CRITICAL', 'ERROR', 'WARNING', 'INFO', 'DEBUG'),
        )

        # global arguments
        for arg, verbose_name in self.global_arguments:
            self.parser.add_argument(
                f'--{arg}', metavar=arg.upper(),
                help=f'Global {verbose_name} argument overwriting context'
            )

        self.db = get_default_db()
        self.logger = logging.getLogger('fuku.client')

    def add_module(self, module):
        module = module(db=self.db, client=self)
        for mod in self.modules:
            if mod.name == module.name:
                raise TypeError('duplicate modules: {}'.format(mod.name))
        self.modules.append(module)

    def add_modules(self, *args):
        for mod in args:
            self.add_module(mod)

    def add_arguments(self):
        subp = self.parser.add_subparsers()

        for mod in self.modules:
            modp = subp.add_parser(mod.name)
            mod.add_arguments(modp)
            modp.set_defaults(handler=mod.entry)

    def get_module(self, name):
        for mod in self.modules:
            if mod.name == name:
                return mod
        raise KeyError('no module named {}'.format(name))

    def get_selected(self, module):
        return self.get_module(module).get_selected()

    def iter_parent_modules(self, name):
        self.logger.debug('Iterating parent modules')
        for mod in self.modules:
            if name in mod.dependencies:
                self.logger.debug(f'\t{mod.name} - dependencies: {mod.dependencies}')
                yield mod

    def iter_dependent_modules(self, parent):
        for mod in self.modules:
            if mod.name in parent.dependencies:
                yield mod

    def entry(self):
        self.add_arguments()
        self.args = self.parser.parse_args()

        # set the logging log level based on parsed arguments
        loglevel = vars(self.args).get('log') or 'WARNING'
        logformat = f'{Fore.CYAN}%(levelname)-10s {Fore.GREEN}%(name)s\t{Fore.RESET}%(message)s'
        logging.basicConfig(level=loglevel, format=logformat)

        try:
            handler = self.args.handler
        except AttributeError:
            handler = None
        if handler:
            handler(self.args)
        save_db(self.db)



import os
import re
import uuid
from datetime import datetime, timedelta
from pprint import pprint

import botocore

from .db import get_rc_path
from .module import Module
from .utils import gen_secret


class Pg(Module):
    dependencies = ['app']

    def __init__(self, **kwargs):
        super().__init__('pg', **kwargs)

    def add_arguments(self, parser):
        # ## SECTION FOR fuku pg ## #
        subp = parser.add_subparsers(help='pg help')

        p = subp.add_parser('ls', help='list postgres DBs')
        p.add_argument('name', metavar='NAME', nargs='?', help='DB name')
        p.set_defaults(pg_handler=self.handle_list)

        p = subp.add_parser('mk', help='add a postgres instance')
        p.add_argument('name', metavar='NAME', help='instance name')
        p.add_argument('--backup', '-b', default=7, type=int, help='number of days to retain backups')
        p.add_argument('--storage', '-s', default=5, type=int, help='allocated storage (GB)')
        p.set_defaults(pg_handler=self.handle_make)

        # p = subp.add_parser('cache', help='cache instance details')
        # p.add_argument('name', metavar='NAME', help='DB name')
        # p.add_argument('password', metavar='PASSWORD', help='DB password')
        # p.set_defaults(pg_handler=self.handle_cache)

        p = subp.add_parser('connect', help='connect to a task')
        p.add_argument('dbname', metavar='DBNAME', help='DB name')
        p.add_argument('--task', '-t', help='target task name')
        p.set_defaults(pg_handler=self.handle_connect)

        p = subp.add_parser('sl', help='select a postgres DB')
        p.add_argument('name', metavar='NAME', nargs='?', help='instance to select')
        p.set_defaults(pg_handler=self.handle_select)

        p = subp.add_parser('psql')
        p.add_argument('--dbname', '-d', help='DB name')
        p.add_argument('--command', '-c', help='run SQL')
        p.set_defaults(pg_handler=self.handle_psql)

        p = subp.add_parser('dump', help='dump contents of database')
        p.add_argument('dbname', metavar='DBNAME', help='DB name')
        p.add_argument('output', metavar='OUTPUT', help='output filename')
        p.set_defaults(pg_handler=self.handle_dump)

        p = subp.add_parser('restore', help='restore a database')
        p.add_argument('dbname', metavar='DBNAME', help='DB name')
        p.add_argument('input', metavar='INPUT', help='database dump file')
        p.set_defaults(pg_handler=self.handle_restore)

        p = subp.add_parser('rollback', help='rollback a database')
        p.add_argument('dbname', metavar='DBNAME', help='DB name')
        p.add_argument('time', metavar='TIME', help='rollback time')
        p.set_defaults(pg_handler=self.handle_rollback)

        p = subp.add_parser('backup', help='backup a database to S3')
        p.add_argument('dbname', metavar='DBNAME', help='DB name')
        p.add_argument('--list', action='store_true', help='list backups')
        p.set_defaults(pg_handler=self.handle_backup)

        p = subp.add_parser('share', help='share a backed up database')
        p.add_argument('dbname', metavar='DBNAME', help='DB name')
        p.add_argument('key', metavar='KEY', help='backup key')
        p.set_defaults(pg_handler=self.handle_share)

        p = subp.add_parser('summary', help='summarize databases')
        p.set_defaults(pg_handler=self.handle_summary)

        # ## SECTION FOR fuku pg db ## #
        p = subp.add_parser('db', help='manage databases')
        ssp = p.add_subparsers()

        p = ssp.add_parser('ls')
        p.set_defaults(pg_handler=self.handle_db_list)

        p = ssp.add_parser('mk')
        p.add_argument('dbname', metavar='DBNAME', help='DB name')
        p.set_defaults(pg_handler=self.handle_db_make)

        p = ssp.add_parser('rm')
        p.add_argument('dbname', metavar='DBNAME', help='DB name')
        p.set_defaults(pg_handler=self.handle_db_remove)

    def handle_list(self, args):
        self.list(args.name)

    def list(self, name):
        rds = self.get_boto_client('rds')
        if name:
            data = rds.describe_db_instances(
                Filters=[{
                    'Name': 'db-instance-id',
                    'Values': [self.get_instance_id(name)]
                }]
            )
            pprint(data['DBInstances'][0])
        else:
            for dbinst in self.iter_db_instances():
                print(dbinst)

    def iter_db_instances(self):
        rds = self.get_boto_client('rds')
        data = rds.describe_db_instances()
        ctx = self.get_context(use_context=False)
        pre = f'fuku-{ctx["cluster"]}-'
        for db in data['DBInstances']:
            name = db['DBInstanceIdentifier']
            if name.startswith(pre):
                yield name[len(pre):]

    def handle_make(self, args):
        self.make(args.name, args.backup, args.storage)

    def make(self, name, backup, storage=5):
        self.use_context = False
        ctx = self.get_context()
        password = gen_secret(16)
        inst_id = self.get_instance_id(name)
        db_id = 'postgres'
        sg_id = self.client.get_module('cluster').get_security_group_id()
        rds = self.get_boto_client('rds')
        rds.create_db_subnet_group(
            DBSubnetGroupName=inst_id,
            DBSubnetGroupDescription=f'Subnet group for {inst_id}',
            SubnetIds=[sn.id for sn in self.get_module('cluster').iter_public_subnets(ctx['cluster'])]
        )
        rds.create_db_instance(
            DBName=db_id,
            DBInstanceIdentifier=inst_id,
            DBInstanceClass='db.t2.micro',
            Engine='postgres',
            AllocatedStorage=storage,
            StorageType='gp2',
            MasterUsername=name,
            MasterUserPassword=password,
            BackupRetentionPeriod=backup,
            DBSubnetGroupName=inst_id,
            VpcSecurityGroupIds=[sg_id],
            PubliclyAccessible=True,
            Tags=[
                {
                    'Key': 'cluster',
                    'Value': ctx['cluster']
                }
            ]
        )
        waiter = rds.get_waiter('db_instance_available')
        waiter.wait(
            DBInstanceIdentifier=inst_id
        )
        self.cache(name, db_id, password)
        self.select(name)

    def handle_db_list(self, args):
        self.db_list()

    def db_list(self):
        for db in self.iter_dbs():
            print(db)

    def iter_dbs(self, inst_name=None):
        if inst_name is None:
            ctx = self.get_context()
            inst_name = ctx['dbinstance']
        else:
            ctx = self.get_context(use_context=False)
        path = os.path.join(ctx['cluster'], f'{inst_name}.pgpass')
        path = self.get_secure_file(path)
        endpoint = self.get_endpoint(inst_name)
        cmd = 'psql -h {} -p {} -U {} -d {}'.format(
            endpoint['Address'],
            endpoint['Port'],
            inst_name,
            'postgres'
        )
        sql = 'SELECT datname FROM pg_database WHERE datistemplate = false;'
        cmd = f'{cmd} -c "{sql}"'
        output = self.run(
            cmd,
            capture=True,
            env={'PGPASSFILE': path}
        )
        for db in output.split()[4:-2]:
            i = db.find('_')
            yield db[i + 1:]

    def handle_db_make(self, args):
        self.db_make(args.dbname)

    def db_make(self, name):
        password = gen_secret(16)
        ctx = self.get_context()
        inst_name = ctx['dbinstance']
        db_id = self.get_db_id(name)
        self.psql(command=f'CREATE DATABASE {db_id} OWNER {inst_name}')
        self.psql(command=f'CREATE ROLE {db_id} NOSUPERUSER NOCREATEDB NOCREATEROLE LOGIN ENCRYPTED PASSWORD \'{password}\'')
        self.psql(command=f'GRANT ALL ON DATABASE {db_id} TO {db_id}')
        self.psql(command=f'GRANT rds_superuser TO {db_id}')
        data = self.get_endpoint(inst_name)
        path = os.path.join(self.get_rc_path(), ctx['app'], inst_name, f'{name}.pgpass')
        try:
            os.makedirs(os.path.dirname(path))
        except OSError:
            pass
        with open(path, 'w') as outf:
            outf.write('{}:{}:{}:{}:{}'.format(
                data['Address'],
                data['Port'],
                db_id,
                db_id,
                password
            ))
        os.chmod(path, 0o600)
        self.encrypt_file(path, purpose='the database credentials')
        s3 = self.get_boto_client('s3')
        s3.upload_file(f'{path}.gpg', ctx['bucket'], f'fuku/{ctx["cluster"]}/{ctx["app"]}/{inst_name}/{name}.pgpass.gpg')

    def handle_db_remove(self, args):
        self.db_remove(args.dbname)

    def db_remove(self, name):
        db_id = self.get_db_id(name)
        self.psql(command=f'DROP DATABASE {db_id}')
        self.psql(command=f'DROP ROLE {db_id}')

    # def handle_cache(self, args):
    #     self.cache(args.name, args.password)

    def cache(self, inst_name, db_name, password):
        ctx = self.get_context()
        data = self.get_endpoint(inst_name)
        path = os.path.join(self.get_rc_path(), f'{inst_name}.pgpass')
        try:
            os.makedirs(os.path.dirname(path))
        except OSError:
            pass
        with open(path, 'w') as outf:
            outf.write('{}:{}:{}:{}:{}'.format(
                data['Address'],
                data['Port'],
                db_name,
                inst_name,
                password
            ))
        os.chmod(path, 0o600)
        self.encrypt_file(path, purpose='the database credentials')
        s3 = self.get_boto_client('s3')
        s3.upload_file(f'{path}.gpg', ctx['bucket'], f'fuku/{ctx["cluster"]}/{inst_name}.pgpass.gpg')

    def handle_connect(self, args):
        self.connect(args.dbname, args.task)

    def connect(self, db_name, task_name):
        task_mod = self.client.get_module('task')
        env = {
            'DATABASE_URL': self.get_url(db_name)
        }
        task_mod.env_set(task_name, env)

    def handle_select(self, args):
        self.select(args.name)

    def select(self, name):
        if name:
            self.use_context = False
            ctx = self.get_context()
            pgpass_path = f'{ctx["cluster"]}/{name}.pgpass'
            # self.exists(name)
            self.get_secure_file(pgpass_path)

        self.store_set('selected', name)
        self.clear_parent_selections()

    def handle_psql(self, args):
        self.psql(args.dbname, args.command)

    def psql(self, db_name=None, command=None):
        ctx = self.get_context()
        inst_name = ctx['dbinstance']
        if db_name:
            db_id, path = self.get_db_creds(db_name)
        else:
            db_id = None
            path = os.path.join(self.get_rc_path(), f'{inst_name}.pgpass')
        endpoint = self.get_endpoint(inst_name)
        cmd = 'psql -h {} -p {} -U {} -d {}'.format(
            endpoint['Address'],
            endpoint['Port'],
            db_id or ctx['dbinstance'],
            db_id or 'postgres'
        )
        if command:
            cmd = f'{cmd} -c "{self.escape(command)}"'
        self.run(
            cmd,
            capture=False,
            env={'PGPASSFILE': path}
        )

    def handle_dump(self, args):
        self.dump(args.dbname, args.output)

    def dump(self, db_name, output):
        ctx = self.get_context()
        db_id, path = self.get_db_creds(db_name)
        endpoint = self.get_endpoint(ctx['dbinstance'])
        self.run(
            'pg_dump -Fc -x -O -h {} -p {} -U {} -d {} -f {}'.format(
                endpoint['Address'],
                endpoint['Port'],
                db_id,
                db_id,
                output
            ),
            capture=False,
            env={'PGPASSFILE': path}
        )

    def handle_restore(self, args):
        self.restore(args.dbname, args.input)

    def restore(self, db_name, input):
        ctx = self.get_context()
        db_id, path = self.get_db_creds(db_name)
        # self.psql(command=f'DROP DATABASE {db_id}')
        # self.psql(command=f'CREATE DATABASE {db_id} OWNER {inst_name}')
        # self.psql(command=f'GRANT ALL ON DATABASE {db_id} TO {db_id}')
        endpoint = self.get_endpoint(ctx['dbinstance'])
        # self.run(
        #     f'psql -h {endpoint["Address"]} -p {endpoint["Port"]} -U {db_name} {db_name} -c \'DROP SCHEMA public CASCADE; CREATE SCHEMA public;\'',
        #     env={'PGPASSFILE': path}
        # )
        self.run(
            f'pg_restore -x -O -c -h {endpoint["Address"]} -p {endpoint["Port"]} -U {db_id} -d {db_id} {input}',
            capture=False,
            env={'PGPASSFILE': path}
        )

    def handle_rollback(self, args):
        self.rollback(args.dbname, args.time)

    def rollback(self, db_name, time_str):
        # ctx = self.get_context()
        db_id, path = self.get_db_creds(db_name)
        # endpoint = self.get_endpoint(ctx['dbinstance'])
        rds_cli = self.get_client('rds')
        m = re.match(r'(\d+):(\d+):(\d+)', time_str)
        if not m:
            self.error('invalid time-ago string')
        time = datetime.utcnow() - timedelta(days=m.groups[1], minutes=m.groups[2], seconds=m.groups[3])
        rds_cli.restore_db_instance_to_point_in_time(
            SourceDBInstanceIdentifier=db_id,
            TargetDBInstanceIdentifier=f'{db_id}_rollback',
            RestoreTime=time
        )

    def handle_backup(self, args):
        self.backup(args.dbname, args.list)

    def backup(self, db_name, list):
        ctx = self.get_context()
        if list:
            pre = 'backups/{}/{}/'.format(
                ctx['dbinstance'],
                db_name,
            )
            for obj in self.iters3(pre):
                key = obj.key[obj.key.rfind('/') + 1:obj.key.rfind('.')]
                print(f'{key}  ({obj.last_modified})')
            return
        db_id, path = self.get_db_creds(db_name)
        endpoint = self.get_endpoint(ctx['dbinstance'])
        cmd = 'pg_dump -Fc -x -O -h {} -p {} -U {} -d {}'.format(
            endpoint['Address'],
            endpoint['Port'],
            db_id,
            db_id
        )
        # key = str(datetime.now()).replace(' ', '-')
        while 1:
            key = str(uuid.uuid4()).replace('-', '')[:8]
            s3 = self.get_boto_resource('s3')
            bucket_key = 'backups/{}/{}/{}.dump'.format(
                ctx['dbinstance'],
                db_name,
                key
            )
            try:
                s3.Object(ctx['bucket'], bucket_key)
                break
            except botocore.exceptions.ClientError as e:
                pass
        cmd += ' | aws s3 cp - s3://{}/backups/{}/{}/{}.dump --profile {}'.format(
            ctx['bucket'],
            ctx['dbinstance'],
            db_name,
            key,
            ctx['profile']
        )
        self.run(
            cmd,
            capture=False,
            env={'PGPASSFILE': path}
        )
        print(f'backed up as "{key}"')

    def handle_share(self, args):
        self.share(args.dbname, args.key)

    def share(self, db_name, key):
        ctx = self.get_context()
        db_id, path = self.get_db_creds(db_name)
        # endpoint = self.get_endpoint(ctx['dbinstance'])
        cmd = 'aws s3 presign s3://{}/backups/{}/{}/{}.dump --profile {}'.format(
            ctx['bucket'],
            ctx['dbinstance'],
            db_name,
            key,
            ctx['profile']
        )
        r = self.run(
            cmd,
            capture=True,
            env={'PGPASSFILE': path}
        )
        print(r)

    def handle_summary(self, args):
        self.summary()

    def summary(self):
        print('')
        for dbinst in self.iter_db_instances():
            print(dbinst)
            for db in self.iter_dbs(dbinst):
                print(f'  {db}')
            print('')

    def get_instance_id(self, instance):
        ctx = self.get_context(use_context=False)
        return f'fuku-{ctx["cluster"]}-{instance}'

    def get_db_id(self, name):
        ctx = self.get_context()
        return f'{ctx["app"]}_{name}'

    def get_rc_path(self):
        ctx = self.get_context(use_context=False)
        return os.path.join(get_rc_path(), ctx['cluster'])

    def get_endpoint(self, name):
        inst_id = self.get_instance_id(name)
        rds = self.get_boto_client('rds')
        try:
            return rds.describe_db_instances(
                DBInstanceIdentifier=inst_id
            )['DBInstances'][0]['Endpoint']
        except:
            self.error(f'no database "{name}"')

    def get_instance(self, inst_name):
        rds = self.get_boto_client('rds')
        return rds.describe_db_instances(
            Filters=[{
                'Name': 'db-instance-id',
                'Values': [self.get_instance_id(inst_name)]
            }]
        )

    def get_url(self, db_name):
        ctx = self.get_context()
        path = os.path.join(self.get_rc_path(), ctx['app'], ctx['dbinstance'], f'{db_name}.pgpass')
        try:
            with open(path, 'r') as inf:
                data = inf.read()
        except:
            self.error(f'no cached information for "{db_name}"')
        host, port, db, user, pw = data.split(':')
        return 'postgres://{}:{}@{}:{}/{}'.format(user, pw, host, port, db)

    # def get_pgpass_file(self, name):
    #     ctx = self.get_context()
    #     path = os.path.join(self.get_rc_path(), '%s.pgpass' % name)
    #     if not os.path.exists(path):
    #         s3 = self.get_boto_client('s3')
    #         s3.download_file(ctx['bucket'], f'fuku/{ctx["cluster"]}/{ctx["app"]}/{inst_name}/{name}.pgpass.gpg', path)
    #         self.run(
    #             'gpg -o {} -d {}.gpg'.format(path, path)
    #         )
    #         os.chmod(path, 0o600)

    def get_db_creds(self, db_name):
        ctx = self.get_context()
        db_id = self.get_db_id(db_name)
        app = ctx['app']
        inst_name = ctx['dbinstance']
        path = f'{ctx["cluster"]}/{app}/{inst_name}/{db_name}.pgpass'
        full_path = self.get_secure_file(path)
        return db_id, full_path

    def get_selected(self, fail=True):
        sel = self.store.get('selected', None)
        if not sel and fail:
            self.error('no postgres DB selected')
        return sel

    def get_my_context(self):
        if self.client.args.pg:
            ctx = {'dbinstance': self.client.args.pg}
        else:
            sel = self.store_get('selected')
            if not sel:
                self.error('no DB currently selected')
            ctx = {'dbinstance': sel}

        self.get_logger().debug(f'PG Context: {ctx}')
        return ctx



from .module import Module


class Route(Module):
    dependencies = ['app']

    def __init__(self, **kwargs):
        super().__init__('route', **kwargs)

    def add_arguments(self, parser):
        subp = parser.add_subparsers(help='route help')

        p = subp.add_parser('mk', help='make a route 53 record set entry')
        p.add_argument('name', metavar='NAME', help='record set name')
        p.add_argument('zone', metavar='ZONE', help='hosted zone name')
        p.set_defaults(route_handler=self.handle_make)

        p = subp.add_parser('rm', help='remove a route 53 record set entry')
        p.add_argument('name', metavar='NAME', help='record set name')
        p.add_argument('zone', metavar='ZONE', help='hosted zone name')
        p.set_defaults(route_handler=self.handle_remove)

    def handle_make(self, args):
        self.make(args.name, args.zone)

    def make(self, name, zone_name):
        # get zone id
        zone = self.get_zone(zone_name)

        # find app target group and elb
        task_mod = self.client.get_module('app')
        target_group = task_mod.get_target_group()

        client = self.get_boto_client('elbv2')
        load_balancers = client.describe_load_balancers(
            LoadBalancerArns=target_group['LoadBalancerArns']
        )['LoadBalancers']
        elb_dns_name = load_balancers[0]['DNSName']

        # map dns name to elb
        client = self.get_boto_client('route53')
        client.change_resource_record_sets(
            HostedZoneId=zone['Id'],
            ChangeBatch= {
                'Changes': [{
                    'Action': 'UPSERT',
                    'ResourceRecordSet': {
                        'Name': f'{name}.{zone_name}.',
                        'Type': 'CNAME',
                        'TTL': 60,
                        'ResourceRecords': [{'Value': elb_dns_name}]
                    }
                }]
            })

    def handle_remove(self, args):
        self.remove(args.name, args.zone)

    def remove(self, name, zone_name):
        # get zone id
        zone = self.get_zone(zone_name)

        record_name = f'{name}.{zone_name}.'
        # make sure record set exists
        client = self.get_boto_client('route53')
        recordset = client.list_resource_record_sets(
            HostedZoneId=zone['Id'],
            StartRecordName=record_name,
            MaxItems='1'
        )['ResourceRecordSets'][0]
        assert recordset['Name'] == record_name

        # delete the recordset
        client.change_resource_record_sets(
            HostedZoneId=zone['Id'],
            ChangeBatch= {
                'Changes': [{
                    'Action': 'DELETE',
                    'ResourceRecordSet': {
                        'Name': record_name,
                        'Type': 'CNAME',
                        'TTL': 60,
                        'ResourceRecords': recordset['ResourceRecords'],
                    }
                }]
            })

    def get_zone(self, zone_name):
        # find dns name
        paginate = self.get_boto_paginator('route53', 'list_hosted_zones').paginate()
        search = paginate.search(f'HostedZones[?Name == `"{zone_name}."`]')
        return list(search)[0]



from .module import Module
from .task import IGNORED_TASK_KWARGS
from .utils import entity_already_exists


class App(Module):
    dependencies = ['cluster']

    def __init__(self, **kwargs):
        super().__init__('app', **kwargs)

    def add_arguments(self, parser):
        subp = parser.add_subparsers(help='app help')

        p = subp.add_parser('ls', help='list applications')
        p.set_defaults(app_handler=self.handle_list)

        p = subp.add_parser('mk', help='add an app')
        p.add_argument('name', metavar='NAME', help='app name')
        p.set_defaults(app_handler=self.handle_make)

        remp = subp.add_parser('rm', help='remove an app')
        remp.add_argument('name', help='app name')
        remp.set_defaults(app_handler=self.handle_remove)

        p = subp.add_parser('sl', help='select an app')
        p.add_argument('name', metavar='NAME', help='app name')
        p.set_defaults(app_handler=self.handle_select)

        p = subp.add_parser('run', help='run a command')
        p.add_argument('image', metavar='IMAGE', help='image name')
        p.add_argument('command', metavar='CMD', nargs='+', help='command to run')
        p.set_defaults(app_handler=self.handle_run)

        p = subp.add_parser('expose', help='expose to internet')
        p.add_argument('name', metavar='NAME', help='app name')
        p.add_argument('domains', metavar='DOMAINS', help='list of domain names', nargs='*')
        p.set_defaults(app_handler=self.handle_expose)

        p = subp.add_parser('hide', help='hide from internet')
        p.add_argument('name', metavar='NAME', help='app name')
        p.set_defaults(app_handler=self.handle_hide)

    def handle_list(self, args):
        self.list()

    def list(self):
        for app in self.iter_apps():
            print(app)

    def iter_apps(self):
        for gr in self.iter_target_groups():
            yield gr['TargetGroupName'].rsplit('-', 1)[1]

    def handle_make(self, args):
        self.make(args.name)

    def make(self, name):
        self.use_context = False

        self.validate(name)

        if self.get_target_group(name):
            self.error(f'App "{name}" already exists')
            return

        self.make_target_group(name)
        self.make_task(name)
        self.select(name)

    def handle_remove(self, args):
        self.remove(args.name)

    def remove(self, name):
        self.use_context = False

        if not self.get_target_group(name):
            self.error(f'No such app "{name}"')
            return

        self.remove_target_group(name)

    def handle_select(self, args):
        self.select(args.name)

    def select(self, name):
        if name is not None:
            for gr in self.iter_target_groups():
                if name and gr['TargetGroupName'].endswith('-' + name):
                    self.store_set('selected', name)
                    self.clear_parent_selections()
                    return
            self.error(f'no app "{name}"')
        else:
            self.clear_parent_selections()

    def handle_run(self, args):
        self.run(args.image, args.command)

    def run(self, img, cmd):
        img = self.client.get_module('image').get_uri(img)
        cmd = ' '.join(cmd or [])
        full_cmd = f'docker run --rm -it {img} {cmd}'
        node_mod = self.client.get_module('node')
        node_mod.mgr_run(full_cmd, tty=True)

    def handle_expose(self, args):
        self.expose(args.name, args.domains)

    def handle_hide(self, args):
        self.hide(args.name)

    def make_task(self, name):
        ctx = self.get_context()
        ecr_cli = self.get_boto_client('ecr')
        with entity_already_exists():
            ecr_cli.create_repository(repositoryName='fuku')

        img_uri = self.client.get_module('image').image_name_to_uri('/fuku')
        task = {
            'family': f'fuku-{ctx["cluster"]}-{name}',
            'containerDefinitions': []
        }
        ctr_def = {
            'name': name,
            'image': img_uri,
            'memoryReservation': 1
        }
        task['containerDefinitions'].append(ctr_def)
        ecs = self.get_boto_client('ecs')
        skip = set(IGNORED_TASK_KWARGS)
        ecs.register_task_definition(**{
            k: v for k, v in task.items() if k not in skip
        })

    def get_my_context(self):
        if self.client.args.app:
            ctx = {'app': self.client.args.app}
        else:
            sel = self.store_get('selected')
            if not sel:
                self.error('no app currently selected')
            ctx = {'app': sel}
        self.get_logger().debug(f'APP Context: {ctx}')
        return ctx


class EcsApp(App):
    def make(self, name):
        super().make(name)
        self.make_target_group(name)

    def make_target_group(self, name):
        ctx = self.get_context()
        vpc_id = self.get_module('cluster').get_vpc(ctx['cluster']).id
        alb_cli = self.get_boto_client('elbv2')
        alb_cli.create_target_group(
            Name=f'fuku-{ctx["cluster"]}-{name}',
            Protocol='HTTP',
            Port=80,
            VpcId=vpc_id,
            Matcher={
                'HttpCode': '200,301'
            }
        )

    def remove_target_group(self, name):
        alb_cli = self.get_boto_client('elbv2')
        target_group = self.get_target_group(name)
        alb_cli.delete_target_group(TargetGroupArn=target_group['TargetGroupArn'])

    def get_target_group(self, app=None):
        ctx = self.get_context()
        alb_cli = self.get_boto_client('elbv2')

        cluster = ctx['cluster']
        if not app:
            app = ctx['app']

        try:
            return alb_cli.describe_target_groups(Names=[f'fuku-{cluster}-{app}', ])['TargetGroups'][0]
        except:
            return None

    def iter_target_groups(self):
        self.use_context = False
        ctx = self.get_context()
        paginate = self.get_boto_paginator('elbv2', 'describe_target_groups').paginate()
        for gr in paginate.search(
            'TargetGroups[?starts_with(TargetGroupName, `"fuku-{cluster}-"`)] '.format(**ctx) +
            '| sort_by(@, &TargetGroupName)'
        ):
            yield gr


    def expose(self, name, domains):
        alb_cli = self.get_boto_client('elbv2')
        target_group = self.get_target_group()

        if target_group['LoadBalancerArns']:
            self.error(
                f"App already exposed on {target_group['LoadBalancerArns']}." +
                '\nPlease remove target group from those load balancers first.'
            )

        # TODO: this is the default limit, look into implementing a check against ec2.describe_account_attributes
        MAX_RULES = 50

        added = False
        paginate = self.get_boto_paginator('elbv2', 'describe_load_balancers').paginate()
        for elb in paginate.search(
            'LoadBalancers[?starts_with(LoadBalancerName, `"fuku-uptick-"`)]'
        ):
            listeners = alb_cli.describe_listeners(LoadBalancerArn=elb['LoadBalancerArn'])
            for listener in listeners['Listeners']:
                rules = alb_cli.describe_rules(ListenerArn=listener['ListenerArn'])['Rules']

                # make sure we have space in the load balancer
                if len(rules) > (MAX_RULES - len(domains)):
                    print(f'{elb["LoadBalancerName"]} [skipped - full]')
                    break

                # count how many rules there are to define next priority
                rule_priorities = [int(r['Priority']) for r in rules if r['Priority'] != 'default']
                priority = (max(rule_priorities) + 1) if rule_priorities else 1

                for i, domain in enumerate(domains):
                    # attempt to create the rule
                    alb_cli.create_rule(
                        ListenerArn=listener['ListenerArn'],
                        Conditions=[
                            {
                                'Field': 'host-header',
                                'Values': [domain]
                            }
                        ],
                        Priority=priority + i,
                        Actions=[
                            {
                                'Type': 'forward',
                                'TargetGroupArn': target_group['TargetGroupArn']
                            }
                        ]
                    )
                    added = True
                    print(f'{elb["LoadBalancerName"]} {domain}:{listener["Port"]} [added]')
            if added:
                break

    def hide(self, name):
        alb_cli = self.get_boto_client('elbv2')

        target_group = self.get_target_group(name)
        if not target_group:
            self.error(f'App {name} is not exposed')

        for load_balancer in target_group['LoadBalancerArns']:
            listeners = alb_cli.describe_listeners(LoadBalancerArn=load_balancer)['Listeners']
            for listener in listeners:
                rules = alb_cli.describe_rules(ListenerArn=listener['ListenerArn'])['Rules']
                for rule in rules:
                    try:
                        if rule['Actions'][0]['TargetGroupArn'] == target_group['TargetGroupArn']:
                            alb_cli.delete_rule(RuleArn=rule['RuleArn'])
                            print(f'Rule on {load_balancer}:{listener["Port"]} [deleted]')
                    except IndexError:
                        pass



import json
import logging
import os
import stat
import sys
import tempfile
import unicodedata
from contextlib import contextmanager
from string import Template

import boto3

from .db import get_rc_path
from .runner import run


class Module(object):
    dependencies = []

    def __init__(self, name, db=None, client=None):
        self.name = name
        self.db = db
        self.store = self.db.setdefault(self.name, {})
        self.client = client
        self.use_context = True
        self._checks = {}

    def add_arguments(self, parser):
        pass

    def validate(self, name):
        for char in {'-', '_', ' ', '/'}:
            if char in name:
                self.error(f'Invalid identifier in name: {unicodedata.name(char)}')

        if name == 'fuku':
            self.error('"fuku" is a reserved name')

    def get_context(self, ctx={}, use_context=True):
        for dep in self.client.iter_dependent_modules(self):
            ctx = dep.get_context(ctx)
        if self.use_context and use_context:
            ctx.update(self.get_my_context())
        return ctx

    def get_module(self, name):
        return self.client.get_module(name)

    def get_selected(self):
        return None

    def error(self, msg):
        print(msg)
        sys.exit()

    def register_check(self, key, call):
        self._checks[key] = call

    def data_path(self, filename=None):
        path = os.path.join(
            os.path.dirname(
                os.path.realpath(__file__)
            ),
            'scripts',
            'data'
        )
        if filename:
            path = os.path.join(path, filename)
        return path

    @contextmanager
    def template_file(self, filename, context={}):
        # with tempfile.NamedTemporaryFile() as outf:
        with open(self.data_path(filename)) as inf:
            data = Template(inf.read()).substitute(context)
        # outf.write(data.encode())
        # outf.flush()
        # yield outf.name
        yield data

    @contextmanager
    def temporary_file(self):
        tf = tempfile.NamedTemporaryFile()
        try:
            yield tf
        finally:
            tf.close()

    def run(self, cmd, cfg={}, capture='discard', use_self=False, ignore_errors=False,
            env={}):
        # cfg = self.merged_config(cfg, use_self)
        # final = subs(cmd, cfg)
        # print(final)
        env_copy = os.environ.copy()
        env_copy.update(env)
        output = run(
            # final,
            cmd,
            capture=capture not in set([None, '', False]),
            ignore_errors=ignore_errors,
            env=env_copy
        )
        if capture == 'json':
            output = json.loads(output)
        return output

    def clear_parent_selections(self):
        self.get_logger().debug(f'Clearing parent selections for {self.name}')
        for parent in self.client.iter_parent_modules(self.name):
            try:
                parent.select(None)
            except AttributeError:
                pass

    def confirm_remove(self, id):
        """ I want to remove something, confirm with parents that this
        is okay.
        """
        results = []
        for parent in self.client.iter_parent_modules(self.name):
            results.extend(parent.dependency_removal(self, id))
        if len(results):
            print('Unable to remove as there are dependencies:\n')
            for res in results:
                mod = res['module'].name
                id = res['id']
                print(f'  module "{mod}", identifier "{id}"')
            print('')

    def dependency_removal(self, module, id):
        """ Something higher up in the tree is going to be removed, check what that
        means for lower modules.
        """
        return []

    def check(self, key):
        if key in self._checks:
            return self._checks[key]()
        else:
            for parent in self.client.iter_parent_modules(self.name):
                res = parent.check(key)
                if res is not None:
                    return res

    def entry(self, args):
        handler = getattr(args, '%s_handler' % self.name, None)
        if handler:
            handler(args)

    def save(self):
        return self.store

    def load(self, cache):
        if 'selected' in cache:
            self.store['selected'] = cache['selected']
        else:
            try:
                del self.store['selected']
            except KeyError:
                pass

    def store_set(self, key, value):
        self.get_logger().debug(f'Set {self.name} store {self.store} with {key}={value}')

        if value:
            self.store[key] = value
            if key == 'selected':
                self.get_logger().info(f'selected: {value}')
        else:
            try:
                del self.store[key]
            except KeyError:
                pass

    def store_get(self, key):
        parts = key.split('.')
        value = self.store
        for p in parts:
            if value is not None:
                value = value.get(p, None)
        return value

    def db_get(self, key):
        parts = key.split('.')
        value = self.store
        for p in parts:
            if value is not None:
                value = value.get(p, None)
        return value

    def setup_boto_session(self, ctx={}):
        ctx = self.get_context(ctx, use_context=False)
        kwargs = {}
        if 'region' in ctx:
            kwargs['region_name'] = ctx['region']
        if 'profile' in ctx:
            kwargs['profile_name'] = ctx['profile']
        boto3.setup_default_session(**kwargs)

    def get_boto_resource(self, resource, ctx={}):
        self.setup_boto_session(ctx)
        return boto3.resource(resource)

    def get_boto_client(self, resource, ctx={}):
        self.setup_boto_session(ctx)
        return boto3.client(resource)

    def get_boto_paginator(self, client, resource, ctx={}):
        self.setup_boto_session(ctx)
        return boto3.client(client).get_paginator(resource)

    def puts3(self, key, value):
        ctx = self.get_context()
        s3 = self.get_boto_client('s3')
        s3.put_object(
            Bucket=ctx['bucket'],
            Key=f'fuku/{key}',
            Body=json.dumps(value)
        )

    def gets3(self, key):
        try:
            ctx = self.get_context(use_context=False)
            s3 = self.get_boto_client('s3')
            data = s3.get_object(
                Bucket=ctx['bucket'],
                Key=f'fuku/{key}',
            )['Body'].read()
            return data
        except:
            return None

    def iters3(self, prefix):
        try:
            ctx = self.get_context(use_context=False)
            s3 = self.get_boto_resource('s3')
            bucket = s3.Bucket(ctx['bucket'])
            for obj in bucket.objects.filter(Prefix=prefix):
                yield obj
        except:
            pass

    def encrypt_file(self, path, purpose='an unknown resource'):
        print(f'\nPlease enter a password to secure {purpose}.')
        print('Be sure to keep this safe, as it is required for future')
        print('access to this resource.\n')
        while 1:
            try:
                self.run('gpg -c {}'.format(path))
                break
            except:
                print('\ntimed out, please try again\n')

    def get_secure_file(self, path):
        full_path = os.path.join(get_rc_path(), path)
        if not os.path.exists(full_path):
            data = self.gets3(f'{path}.gpg')
            if data is None:
                self.error(f'no secure key file found: {path}')
            try:
                os.makedirs(os.path.dirname(full_path))
            except OSError:
                pass
            with open(f'{full_path}.gpg', 'wb') as file:
                file.write(data)
            try:
                self.run(f'gpg -d {full_path}.gpg > {full_path}')
            except:
                self.clear_secure_file(path)
                raise
            os.chmod(full_path, stat.S_IRUSR | stat.S_IWUSR)
        return full_path

    def clear_secure_file(self, path):
        full_path = os.path.join(get_rc_path(), path)
        try:
            os.remove(full_path)
        except:
            pass

    def get_my_context(self):
        return {}

    def escape(self, value):
        return value.replace('"', '\\"')

    def get_logger(self):
        logger = logging.getLogger(f'fuku.{self.name}')
        return logger
